# -*- coding: utf-8 -*-

from __future__ import print_function, division

"""
Specification for various versions of AVM
"""

from copy import deepcopy

from .datatypes import *
from .cv import *

SPECS = {}

SPECS[1.1] = {

    # Creator Metadata
    'Creator': AVMString('photoshop:Source'),
    'CreatorURL': AVMURL('Iptc4xmpCore:CreatorContactInfo.CiUrlWork'),
    'Contact.Name': AVMOrderedList('dc:creator'),
    'Contact.Email': AVMEmail('Iptc4xmpCore:CreatorContactInfo.CiEmailWork'),
    'Contact.Address': AVMString('Iptc4xmpCore:CreatorContactInfo.CiAdrExtadr'),
    'Contact.Telephone': AVMString('Iptc4xmpCore:CreatorContactInfo.CiTelWork'),
    'Contact.City': AVMString('Iptc4xmpCore:CreatorContactInfo.CiAdrCity'),
    'Contact.StateProvince': AVMString('Iptc4xmpCore:CreatorContactInfo.CiAdrRegion'),
    'Contact.PostalCode': AVMString('Iptc4xmpCore:CreatorContactInfo.CiAdrPcode'),
    'Contact.Country': AVMString('Iptc4xmpCore:CreatorContactInfo.CiAdrCtry'),
    'Rights': AVMLocalizedString('xapRights:UsageTerms'),

    # Content Metadata
    'Title': AVMLocalizedString('dc:title'),
    'Headline': AVMString('photoshop:Headline'),
    'Description': AVMLocalizedString('dc:description'),
    'Subject.Category': AVMUnorderedStringList('avm:Subject.Category'),
    'Subject.Name': AVMUnorderedStringList('dc:subject'),
    'Distance': AVMOrderedFloatList('avm:Distance', length=2, strict_length=False),
    'Distance.Notes': AVMString('avm:Distance.Notes'),
    'ReferenceURL': AVMURL('avm:ReferenceURL'),
    'Credit': AVMString('photoshop:Credit'),
    'Date': AVMDateTime('photoshop:DateCreated'),
    'ID': AVMString('avm:ID'),
    'Type': AVMStringCVCapitalize('avm:Type', TYPE_CHOICES),
    'Image.ProductQuality': AVMStringCVCapitalize('avm:Image.ProductQuality', IMAGE_PRODUCT_QUALITY_CHOICES),

    # Observation Metadata
    'Facility': AVMOrderedList('avm:Facility'),
    'Instrument': AVMOrderedList('avm:Instrument'),
    'Spectral.ColorAssignment': AVMOrderedListCV('avm:Spectral.ColorAssignment', SPECTRAL_COLOR_ASSIGNMENT_CHOICES),
    'Spectral.Band': AVMOrderedListCV('avm:Spectral.Band', SPECTRAL_BAND_CHOICES),
    'Spectral.Bandpass': AVMOrderedList('avm:Spectral.Bandpass'),
    'Spectral.CentralWavelength': AVMOrderedFloatList('avm:Spectral.CentralWavelength'),
    'Spectral.Notes': AVMLocalizedString('avm:Spectral.Notes'),
    'Temporal.StartTime': AVMDateTimeList('avm:Temporal.StartTime'),
    'Temporal.IntegrationTime': AVMOrderedFloatList('avm:Temporal.IntegrationTime'),
    'DatasetID': AVMOrderedList('avm:DatasetID'),

    # Coordinate Metadata
    'Spatial.CoordinateFrame': AVMStringCVUpper('avm:Spatial.CoordinateFrame', SPATIAL_COORDINATE_FRAME_CHOICES),
    'Spatial.Equinox': AVMString('avm:Spatial.Equinox'),
    'Spatial.ReferenceValue': AVMOrderedFloatList('avm:Spatial.ReferenceValue', length=2, strict_length=True),
    'Spatial.ReferenceDimension': AVMOrderedFloatList('avm:Spatial.ReferenceDimension', length=2, strict_length=True),
    'Spatial.ReferencePixel': AVMOrderedFloatList('avm:Spatial.ReferencePixel', length=2, strict_length=True),
    'Spatial.Scale': AVMOrderedFloatList('avm:Spatial.Scale', length=2, strict_length=True),
    'Spatial.Rotation': AVMFloat('avm:Spatial.Rotation'),
    'Spatial.CoordsystemProjection': AVMStringCVUpper('avm:Spatial.CoordsystemProjection', SPATIAL_COORDSYSTEM_PROJECTION_CHOICES),
    'Spatial.Quality': AVMStringCVCapitalize('avm:Spatial.Quality', SPATIAL_QUALITY_CHOICES),
    'Spatial.Notes': AVMLocalizedString('avm:Spatial.Notes'),
    'Spatial.FITSheader': AVMString('avm:Spatial.FITSheader'),
    'Spatial.CDMatrix': AVMOrderedFloatList('avm:Spatial.CDMatrix', length=4, strict_length=True, deprecated=True),

    # Publisher Metadata
    'Publisher': AVMString('avm:Publisher'),
    'PublisherID': AVMString('avm:PublisherID'),
    'ResourceID': AVMString('avm:ResourceID'),
    'ResourceURL': AVMURL('avm:ResourceURL'),
    'RelatedResources': AVMUnorderedStringList('avm:RelatedResources'),
    'MetadataDate': AVMDateTime('avm:MetadataDate'),
    'MetadataVersion': AVMFloat('avm:MetadataVersion'),

    # FITS Liberator Metadata

    'FL.BackgroundLevel': AVMOrderedFloatList('avm:FL.BackgroundLevel'),
    'FL.BlackLevel': AVMOrderedFloatList('avm:FL.BlackLevel'),
    'FL.ScaledPeakLevel': AVMOrderedFloatList('avm:FL.ScaledPeakLevel'),
    'FL.PeakLevel': AVMOrderedFloatList('avm:FL.PeakLevel'),
    'FL.WhiteLevel': AVMOrderedFloatList('avm:FL.WhiteLevel'),
    'FL.ScaledBackgroundLevel': AVMOrderedFloatList('avm:FL.ScaledBackgroundLevel'),
    'FL.StretchFunction': AVMOrderedList('avm:FL.StretchFunction')
}

# TODO: write specification for version 1.0
SPECS[1.0] = deepcopy(SPECS[1.1])

SPECS[1.2] = deepcopy(SPECS[1.1])

# Content Metadata

SPECS[1.2]['PublicationID'] = AVMUnorderedStringList('avm:PublicationID')
SPECS[1.2]['ProposalID'] = AVMUnorderedStringList('avm:ProposalID')
SPECS[1.2]["RelatedResources"] = AVMUnorderedStringList('avm:RelatedResources', deprecated=True)

# Create reverse lookup

REVERSE_SPECS = {}
for spec in SPECS:
    REVERSE_SPECS[spec] = {}
    for key in SPECS[spec]:
        value = SPECS[spec][key]
        REVERSE_SPECS[spec][value.namespace, value.tag] = key



# PyAVM - Simple pure-python AVM meta-data handling
# Copyright (c) 2011-13 Thomas P. Robitaille
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from __future__ import print_function, division

try:
    unicode
except:
    basestring = unicode = str

import warnings
try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO
import xml.etree.ElementTree as et

from .specs import SPECS, REVERSE_SPECS


def register_namespace(tag, uri):
    try:
        et.register_namespace(tag, uri)
    except:
        et._namespace_map[uri] = tag

try:
    from astropy.wcs import WCS
    from astropy.io import fits
    astropy_installed = True
except ImportError:
    astropy_installed = False


class NoSpatialInformation(Exception):
    pass


from .embed import embed_xmp
from .extract import extract_xmp

# Define namespace to tag mapping

namespaces = {}
namespaces['http://www.communicatingastronomy.org/avm/1.0/'] = 'avm'
namespaces['http://iptc.org/std/Iptc4xmpCore/1.0/xmlns/'] = 'Iptc4xmpCore'
namespaces['http://purl.org/dc/elements/1.1/'] = 'dc'
namespaces['http://ns.adobe.com/photoshop/1.0/'] = 'photoshop'
namespaces['http://ns.adobe.com/xap/1.0/rights/'] = 'xapRights'

reverse_namespaces = {}
for key in namespaces:
    reverse_namespaces[namespaces[key]] = key


class NoAVMPresent(Exception):
    pass


def capitalize(string):
    return string[0].upper() + string[1:]


def utf8(value):
    return unicode(value).encode('utf-8')


def decode_ascii(string):
    try:
        return string.decode('ascii')
    except AttributeError:  # already a string
        return string


def auto_type(string):
    """
    Try and convert a string to an integer or float
    """
    try:
        return int(string)
    except:
        try:
            return float(string)
        except:
            return string


class AVMContainer(object):

    def __init__(self, allow_value=False):
        if allow_value:
            self.value = None
        self._items = {}

    def __str__(self, indent=0):

        string = ""
        for family in self._items:

            if family.startswith('_'):
                continue

            if type(self._items[family]) is AVMContainer:
                substring = self._items[family].__str__(indent + 3)
                if substring != "":
                    if hasattr(self._items[family], 'value'):
                        string += indent * " " + "%s: %s\n" % (family, utf8(self._items[family].value))
                    else:
                        string += indent * " " + "%s:\n" % family
                    string += substring
            else:
                if type(self._items[family]) is list:
                    string += indent * " " + "%s:\n" % family
                    for elem in self._items[family]:
                        if elem is not None:
                            string += indent * " " + "   * %s\n" % utf8(elem)
                else:
                    if self._items[family] is not None:
                        string += indent * " " + \
                            "%s: %s\n" % (family, utf8(self._items[family]))

        return string

    def __repr__(self):
        return self.__str__()

    def __setattr__(self, attribute, value):
        if attribute in ['_items', 'value']:
            object.__setattr__(self, attribute, value)
            return
        if attribute not in self._items:
            raise Exception("%s is not a valid AVM tag" % attribute)
        else:
            self._items[attribute] = value

    def __getattr__(self, attribute):
        if attribute in self._items:
            return self._items[attribute]
        else:
            return object.__getattr__(self, attribute)


def parse_avm_content(rdf):

    avm_content = {}

    for item in rdf.attrib:

        # Find URI
        uri, tag = item[1:].split('}')

        if uri in namespaces:
            avm_content[(namespaces[uri], tag)] = rdf.attrib[item]

    for item in rdf:

        # Find URI
        uri, tag = item.tag[1:].split('}')

        if uri == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#':
            sub_avm_content = parse_avm_content(item)
            for key in sub_avm_content:
                avm_content[key] = sub_avm_content[key]
        elif uri in namespaces:
            if len(item) == 0:
                avm_content[(namespaces[uri], tag)] = item.text
            elif len(item) == 1:
                c_uri, c_tag = item[0].tag[1:].split('}')
                if c_uri == 'http://www.w3.org/1999/02/22-rdf-syntax-ns#' and c_tag in ['Bag', 'Seq', 'Alt']:
                    avm_content[(namespaces[uri], tag)] = [x.text for x in item[0]]
                else:
                    raise Exception("Unexpected tag %s:%s" % (c_uri, c_tag))
            elif len(item) > 1:
                sub_avm_content = parse_avm_content(item)
                for key in sub_avm_content:
                    avm_content[key] = sub_avm_content[key]

    return avm_content


class AVM(AVMContainer):
    """
    There are several ways to initialize an AVM object:

    * Initialize an empty AVM object:

        >>> avm = AVM()

    * Parse AVM meta-data from an existing image:

        >>> avm = AVM.from_image('myexample.jpg')

    * Create an AVM object from a FITS header:

        >>> from astropy.io import fits
        >>> header = fits.getheader('image.fits')
        >>> avm = AVM.from_header(header)

    * Create an AVM meta-data object from an Astropy WCS instance:

        >>> from astropy.wcs import WCS
        >>> from pyavm import AVM
        >>> wcs = WCS('image.fits')
        >>> avm = AVM.from_wcs(wcs)

    View the contents of the AVM object:

        >>> print(avm)

    The AVM meta-data can be accessed using the attribute notation:

        >>> avm.Spatial.Equinox
        'J2000'
        >>> avm.Publisher
        'Chandra X-ray Observatory'

    Tags can be modified:

        >>> avm.Spatial.Equinox = "B1950"
        >>> avm.Spatial.Notes = "The WCS information was updated on 04/02/2010"

    Finally, it is possible to embed AVM meta-data into an image file:

        >>> avm.embed('original_image.jpg', 'tagged_image.jpg')

    At this time, only JPG and PNG files are supported.
    """

    def __init__(self, origin=None, version=1.2):

        self._items = {}

        self.MetadataVersion = version

        self._update_attributes()

    def _update_attributes(self):

        # Remove attributes that are no longer in the specs

        remove_base = []
        for avm_name in self._items:
            item = self._items[avm_name]
            if isinstance(item, AVMContainer):
                remove_sub = []
                for avm_subset in item._items:
                    full_name = '{0}.{1}'.format(avm_name, avm_subset)
                    if not full_name in self._specs:
                        if item._items[avm_subset] is not None:
                            warnings.warn("{0} is not defined in format specification {1} and will be deleted".format(full_name, self.MetadataVersion))
                        remove_sub.append(avm_subset)
                for key in remove_sub:
                    item._items.pop(key)
            else:
                if not avm_name in self._specs:
                    if self._items[avm_name] is not None:
                        warnings.warn("{0} is not defined in format specification {1} and will be deleted".format(avm_name, self.MetadataVersion))
                    remove_base.append(avm_name)
        for key in remove_base:
            self._items.pop(key)

        # Add any missing ones

        # First pass for root-level attributes, second pass for nested ones
        for iteration in range(2):
            for avm_name in self._specs:

                if avm_name == "MetadataVersion":
                    continue

                if "Distance" in avm_name:
                    if not "Distance" in self._items:
                        self._items['Distance'] = AVMContainer(allow_value=True)

                if iteration == 0 and not '.' in avm_name:

                    if avm_name not in self._items:
                        if "Distance" in avm_name:
                            self._items[avm_name] = AVMContainer(allow_value=True)
                        else:
                            self._items[avm_name] = None

                elif iteration == 1 and '.' in avm_name:

                    family, key = avm_name.split('.')
                    if not family in self._items:
                        self._items[family] = AVMContainer()
                    if not key in self._items[family]._items:
                        self._items[family]._items[key] = None

    def __dir__(self):
        attributes = []
        for key in self._items:
            if '.' in key:
                attribute = key.split('.')[0]
            else:
                attribute = key
            if not attribute in attributes:
                attributes.append(attribute)
        return attributes

    @property
    def _specs(self):
        return SPECS[self.MetadataVersion]

    @property
    def _reverse_specs(self):
        return REVERSE_SPECS[self.MetadataVersion]

    @property
    def MetadataVersion(self):
        if 'MetadataVersion' in self._items:
            return self._items['MetadataVersion']
        else:
            return None

    @MetadataVersion.setter
    def MetadataVersion(self, value):
        self._items['MetadataVersion'] = value
        self._update_attributes()

    def __setattr__(self, attribute, value):

        if attribute in ['_items', 'MetadataVersion']:
            object.__setattr__(self, attribute, value)
            return

        if attribute not in self._specs:
            raise AttributeError("{0} is not a valid AVM group or tag in the {1} standard".format(attribute, self.MetadataVersion))

        avm_class = self._specs[attribute]
        value = avm_class.check_data(value)

        if attribute in self._items and isinstance(self._items[attribute], AVMContainer):
            if hasattr(self._items[attribute], "value"):
                self._items[attribute].value = value
            else:
                raise AttributeError("{0} is an AVM group, not a tag".format(attribute))
        else:
            self._items[attribute] = value

    def __getattr__(self, attribute):
        if attribute in self._items:
            return self._items[attribute]
        else:
            return object.__getattr__(self, attribute)

    @classmethod
    def from_image(cls, filename):
        """
        Instantiate an AVM object from an existing image.
        """

        # Get XMP data from file
        xmp = extract_xmp(filename)

        # Extract XML
        start = xmp.index("<?xpacket begin=")
        start = xmp.index("?>", start) + 2
        end = xmp.index("</x:xmpmeta>") + 12

        # Extract XML
        xml = xmp[start:end]
        return cls.from_xml(xml)

    @classmethod
    def from_xml_file(cls, filename):
        """
        Instantiate an AVM object from an xml file.
        """
        return cls.from_xml(open(filename, 'rb').read())

    @classmethod
    def from_xml(cls, xml):
        """
        Instantiate an AVM object from an XML string
        """

        self = cls()

        # Parse XML
        tree = et.parse(StringIO(xml))
        root = tree.getroot()
        avm_content = parse_avm_content(root)

        for tag, name in avm_content:

            content = avm_content[(tag, name)]

            if (tag, name) in self._reverse_specs:

                avm_name = self._reverse_specs[tag, name]

                # Add to AVM dictionary
                avm_class = self._specs[avm_name]
                content = avm_class.check_data(content)
                if "." in avm_name:
                    family, key = avm_name.split('.')
                    self._items[family]._items[key] = content
                else:
                    if hasattr(self._items[avm_name], 'value'):
                        self._items[avm_name].value = content
                    else:
                        self._items[avm_name] = content

            else:

                warnings.warn("ignoring tag %s:%s" % (tag, name))

        return self

    def to_wcs(self, use_full_header=False, target_image=None):
        """
        Convert AVM projection information into a Astropy WCS object.

        Parameters
        ----------
        use_full_header : bool, optional
            Whether to use the full embedded Header if available. If set to
            `False`, the WCS is determined from the regular AVM keywords.
        target_image : str, optional
            In some cases, the dimensions of the image containing the AVM/WCS
            information is different from the dimensions of the image for which
            the AVM was defined. The `target_image` option can be used to pass
            the path of an image from which the size will be used to re-scale
            the WCS.
        """

        if not astropy_installed:
            raise Exception("Astropy is required to use to_wcs()")

        if repr(self.Spatial) == '':
            raise NoSpatialInformation("AVM meta-data does not contain any spatial information")

        if use_full_header and self.Spatial.FITSheader is not None:
            print("Using full FITS header from Spatial.FITSheader")
            header = fits.Header(txtfile=StringIO(self.Spatial.FITSheader))
            return WCS(header)

        # Initializing WCS object
        wcs = WCS(naxis=2)

        # Find the coordinate type
        if self.Spatial.CoordinateFrame is not None:
            ctype = self.Spatial.CoordinateFrame
        else:
            warnings.warn("Spatial.CoordinateFrame not found, assuming ICRS")
            ctype = 'ICRS'

        if ctype in ['ICRS', 'FK5', 'FK4']:
            xcoord = "RA--"
            ycoord = "DEC-"
            wcs.wcs.radesys = ctype.encode('ascii')
        elif ctype in ['ECL']:
            xcoord = "ELON"
            ycoord = "ELAT"
        elif ctype in ['GAL']:
            xcoord = "GLON"
            ycoord = "GLAT"
        elif ctype in ['SGAL']:
            xcoord = "SLON"
            ycoord = "SLAT"
        else:
            raise Exception("Unknown coordinate system: %s" % ctype)

        # Find the projection type
        cproj = ('%+4s' % self.Spatial.CoordsystemProjection).replace(' ', '-')

        wcs.wcs.ctype[0] = (xcoord + cproj).encode('ascii')
        wcs.wcs.ctype[1] = (ycoord + cproj).encode('ascii')

        # Find the equinox
        if self.Spatial.Equinox is None:
            warnings.warn("Spatial.Equinox is not present, assuming 2000")
            wcs.wcs.equinox = 2000.
        elif type(self.Spatial.Equinox) is str:
            if self.Spatial.Equinox == "J2000":
                wcs.wcs.equinox = 2000.
            elif self.Spatial.Equinox == "B1950":
                wcs.wcs.equinox = 1950.
            else:
                try:
                    wcs.wcs.equinox = float(self.Spatial.Equinox)
                except ValueError:
                    raise ValueError("Unknown equinox: %s" % self.Spatial.Equinox)
        else:
            wcs.wcs.equinox = float(self.Spatial.Equinox)

        # Set standard WCS parameters
        if self.Spatial.ReferenceDimension is not None:
            wcs_naxis1, wcs_naxis2 = self.Spatial.ReferenceDimension
            if hasattr(wcs, 'naxis1'):  # PyWCS and Astropy < 0.4
                wcs.naxis1, wcs.naxis2 = wcs_naxis1, wcs_naxis2
        else:
            wcs_naxis1, wcs_naxis2 = None, None

        wcs.wcs.crval = self.Spatial.ReferenceValue
        wcs.wcs.crpix = self.Spatial.ReferencePixel

        if self.Spatial.CDMatrix is not None:
            wcs.wcs.cd = [self.Spatial.CDMatrix[0:2],
                          self.Spatial.CDMatrix[2:4]]
        elif self.Spatial.Scale is not None:
            # AVM Standard 1.2:
            #
            # "The scale should follow the standard FITS convention for sky
            # projections in which the first element is negative (indicating
            # increasing RA/longitude to the left) and the second is positive.
            # In practice, only the absolute value of the first term should be
            # necessary to identify the pixel scale since images should always
            # be presented in an undistorted 1:1 aspect ratio as they appear in
            # the sky when viewed from Earth.This field can be populated from
            # the FITS keywords: CDELT1, CDELT2 (or derived from CD matrix)."
            #
            # Therefore, we have to enforce the sign of CDELT:
            wcs.wcs.cdelt[0] = - abs(self.Spatial.Scale[0])
            wcs.wcs.cdelt[1] = + abs(self.Spatial.Scale[1])
            if self.Spatial.Rotation is not None:
                wcs.wcs.crota = self.Spatial.Rotation, self.Spatial.Rotation

        # If `target_image` is set, we have to rescale the reference pixel and
        # the scale
        if target_image is not None:

            # Find target image size
            from PIL import Image
            nx, ny = Image.open(target_image).size

            if self.Spatial.ReferenceDimension is None:
                raise ValueError("Spatial.ReferenceDimension should be set in order to determine scale in target image")

            # Find scale in x and y
            scale_x = nx / float(wcs_naxis1)
            scale_y = ny / float(wcs_naxis2)

            # Check that scales are consistent
            if abs(scale_x - scale_y) / (scale_x + scale_y) * 2. < 0.01:
                scale = scale_x
            else:
                raise ValueError("Cannot scale WCS to target image consistently in x and y direction")

            wcs.wcs.cdelt /= scale
            wcs.wcs.crpix *= scale

            if hasattr(wcs, 'naxis1'):  # PyWCS and Astropy < 0.4
                wcs.naxis1 = nx
                wcs.naxis2 = ny

        return wcs

    @classmethod
    def from_header(cls, header, include_full_header=True):
        """
        Instantiate an AVM object from a FITS header
        """

        if not astropy_installed:
            raise Exception("Astropy is required to use from_wcs()")

        wcs = WCS(header)
        shape = (header['NAXIS2'], header['NAXIS1'])
        self = cls.from_wcs(wcs, shape=shape)

        if include_full_header:
            self.Spatial.FITSheader = str(header)

        return self

    @classmethod
    def from_wcs(cls, wcs, shape=None):
        """
        Instantiate an AVM object from a WCS transformation

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS` instance
            The WCS to convert to AVM
        shape : tuple, optional
            The shape of the image (using Numpy y, x order)
        """

        if not astropy_installed:
            raise Exception("Astropy is required to use from_wcs()")

        self = cls()

        # Equinox

        self.Spatial.Equinox = wcs.wcs.equinox

        # Projection

        proj1 = wcs.wcs.ctype[0][-3:]
        proj2 = wcs.wcs.ctype[1][-3:]
        if proj1 == proj2:
            self.Spatial.CoordsystemProjection = decode_ascii(proj1)
        else:
            raise Exception("Projections do not agree: %s / %s" % (proj1, proj2))

        try:
            self.Spatial.ReferenceDimension = [wcs.naxis1, wcs.naxis2]
        except:
            if shape is None:
                warnings.warn("no shape specified, so Spatial.ReferenceDimension will not be set")
            else:
                self.Spatial.ReferenceDimension = [shape[1], shape[0]]

        self.Spatial.ReferenceValue = wcs.wcs.crval.tolist()
        self.Spatial.ReferencePixel = wcs.wcs.crpix.tolist()

        # The following is required to cover all cases of CDELT/PC/CD
        import numpy as np  # will be installed if Astropy is present
        cdelt = np.matrix(wcs.wcs.get_cdelt())
        pc = np.matrix(wcs.wcs.get_pc())
        scale = np.array(cdelt * pc)[0,:].tolist()
        self.Spatial.Scale = scale

        xcoord = decode_ascii(wcs.wcs.ctype[0][:4])
        ycoord = decode_ascii(wcs.wcs.ctype[1][:4])

        if xcoord == 'RA--' and ycoord == 'DEC-':
            if wcs.wcs.radesys in ('ICRS', 'FK5', 'FK4'):
                self.Spatial.CoordinateFrame = str(wcs.wcs.radesys)
            else:  # assume epoch-independent coordinate system
                warnings.warn("RADESYS header keyword not found, assuming ICRS")
                self.Spatial.CoordinateFrame = 'ICRS'
        elif xcoord == 'ELON' and ycoord == 'ELAT':
            self.Spatial.CoordinateFrame = 'ECL'
        elif xcoord == 'GLON' and ycoord == 'GLAT':
            self.Spatial.CoordinateFrame = 'GAL'
        elif xcoord == 'SLON' and ycoord == 'SLAT':
            self.Spatial.CoordinateFrame = 'SGAL'
        else:
            raise Exception("Unknown coordinate system: {0}/{1}".format(xcoord, ycoord))

        try:
            self.Spatial.Rotation = wcs.wcs.crota[1]
        except:
            pass

        self.Spatial.Quality = "Full"

        return self

    def to_xml(self):
        """
        Convert the AVM meta-data to an XML string
        """

        # Register namespaces
        register_namespace('x', "adobe:ns:meta/")
        register_namespace('rdf', 'http://www.w3.org/1999/02/22-rdf-syntax-ns#')
        for namespace in namespaces:
            register_namespace(namespaces[namespace], namespace)

        # Create containing structure
        root = et.Element("{adobe:ns:meta/}xmpmeta")
        trunk = et.SubElement(root, "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}RDF")
        branch = et.SubElement(trunk, "{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description")

        self.MetadataVersion = 1.1

        # Write all the elements
        for name in self._items:
            if isinstance(self._items[name], AVMContainer):
                for key in self._items[name]._items:
                    if self._items[name]._items[key] is not None:
                        if key == "value":
                            avm_class = self._specs['%s' % name]
                            avm_class.to_xml(branch, self._items[name].value)
                        else:
                            avm_class = self._specs['%s.%s' % (name, key)]
                            avm_class.to_xml(branch, self._items[name]._items[key])
            else:
                if self._items[name] is not None and name in self._specs:
                    avm_class = self._specs[name]
                    avm_class.to_xml(branch, self._items[name])

        # Create XML Tree
        tree = et.ElementTree(root)

        # Need to create a StringIO object to write to
        s = StringIO()
        tree.write(s, encoding='utf-8')

        # Rewind and read the contents
        s.seek(0)
        xml_string = s.read()

        return xml_string

    def to_xmp(self):
        """
        Convert the AVM meta-data to an XMP packet
        """
        packet = b'<?xpacket begin="\xef\xbb\xbf" id="W5M0MpCehiHzreSzNTczkc9d"?>\n'
        packet += self.to_xml()
        packet += b'<?xpacket end="w"?>'

        return packet

    def embed(self, filename_in, filename_out, verify=False):
        """
        Embed the AVM meta-data in an image file
        """

        # Embed XMP packet into file
        embed_xmp(filename_in, filename_out, self.to_xmp())

        # Verify file if needed
        if verify:
            try:
                from PIL import Image
            except ImportError:
                try:
                    import Image
                except ImportError:
                    raise ImportError("PIL is required for the verify= option")
            image = Image.open(filename_out)
            image.verify()



from __future__ import print_function, division

try:
    unicode
except:
    basestring = unicode = str

import re
import datetime
import warnings
import xml.etree.ElementTree as et

from .exceptions import AVMItemNotInControlledVocabularyError, AVMListLengthError

__all__ = [
    'AVMString',
    'AVMStringCVCapitalize',
    'AVMStringCVUpper',
    'AVMURL',
    'AVMEmail',
    'AVMLocalizedString',
    'AVMFloat',
    'AVMUnorderedStringList',
    'AVMOrderedList',
    'AVMOrderedListCV',
    'AVMOrderedFloatList',
    'AVMDate',
    'AVMDateTime',
    'AVMDateTimeList',
]


namespaces = {}
namespaces['http://www.communicatingastronomy.org/avm/1.0/'] = 'avm'
namespaces['http://iptc.org/std/Iptc4xmpCore/1.0/xmlns/'] = 'Iptc4xmpCore'
namespaces['http://purl.org/dc/elements/1.1/'] = 'dc'
namespaces['http://ns.adobe.com/photoshop/1.0/'] = 'photoshop'
namespaces['http://ns.adobe.com/xap/1.0/rights/'] = 'xapRights'

reverse_namespaces = {}
for key in namespaces:
    reverse_namespaces[namespaces[key]] = key


class AVMData(object):
    """
    Abstract AVM data class.  All other data classes inherit from AVMData.
    """

    def __init__(self, path, deprecated=False, **kwargs):
        """ """
        self.namespace, self.tag = path.split(':')
        self.deprecated = deprecated

    def check_data(self, value):
        """
        All other data classes should define check_data() based on the type of data.
        Encoding of string into UTF-8 happens here.

        :return: String (UTF-8)
        """
        return value


class AVMString(AVMData):
    """
    Data type for strings
    """

    def check_data(self, value):
        """
        Check that the data is a string or unicode, otherwise it raises a TypeError.

        :return: String (UTF-8)
        """
        if not value:
            return None
        if isinstance(value, (list, tuple)) and len(value) == 1:
            value = value[0]
        if isinstance(value, basestring):
            return value
        elif value is None:
            return None
        else:
            raise TypeError("{0:s} is not a string or unicode".format(self.tag))

    def to_xml(self, parent, value):
        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))
        element.text = "%s" % value
        return element


class AVMURL(AVMString):
    """
    Data type for URLs.

    :return: String (UTF-8)
    """

    def check_data(self, value):
        """
        Checks the data is a string or unicode, and checks data
        against a regular expression for a URL.  If the user leaves
        off the protocol,then 'http://' is attached as a default.

        :return: String (UTF-8)
        """
        if not value:
            return None

        if not (isinstance(value, basestring)):
            raise TypeError("{0:s} is not a string or unicode".format(self.tag))

        value = value

        if value and '://' not in value:
            value = 'http://%s' % value

        url_re = re.compile(
            r'^https?://'  # http:// or https://
            r'(?:(?:[A-Z0-9-]+\.)+[A-Z]{2,6}|'  # domain...
            r'localhost|'  # localhost...
            r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
            r'(?::\d+)?'  # optional port
            r'(?:/?|/\S+)$', re.IGNORECASE)

        if not re.search(url_re, value):
            warnings.warn("{0:s} is not a valid URL".format(self.tag))

        return value


class AVMEmail(AVMString):
    """
    Data type for email addresses.

    :return: String (UTF-8)
    """

    def check_data(self, value):
        """
        Checks data is a string or unicode, and checks against a regular expression
        for an email.  If value is not a string or unicode, a TypeError is raised.
        If the value is not a proper email, then a ValueError is raised.

        :return: String (UTF-8)
        """
        if not (isinstance(value, basestring)):
            raise TypeError("{0:s} is not a string or unicode".format(self.tag))

        value = value

        email_re = re.compile(
            r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"  # dot-atom
            r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-011\013\014\016-\177])*"'  # quoted-string
            r')@(?:[A-Z0-9-]+\.)+[A-Z]{2,6}$', re.IGNORECASE
        )

        if not re.search(email_re, value):
            warnings.warn("{0:s} is not a valid email address".format(self.tag))

        return value


class AVMStringCV(AVMString):
    """ """

    def __init__(self, path, cv, **kwargs):
        self.controlled_vocabulary = cv
        super(AVMStringCV, self).__init__(path, **kwargs)

    def format_data(self, value):
        """
        :return: String
        """
        return value

    def check_cv(self, value):
        """
        If a controlled vocabulary is specified, this function checks the input value against the allowed values.
        AVMItemNotInControlledVocabularyError is raised if not in the controlled vocabulary

        :return: Boolean
        """
        if value in self.controlled_vocabulary:
            return True
        else:
            return False

    def check_data(self, value):
        """
        Check that the data is a string or unicode, formats the data appropriately using format_data()
        and calls check_cv()

        :return: String (UTF-8)
        """
        if not value:
            return None

        if isinstance(value, basestring):
            value = value
            value = self.format_data(value)

            if self.check_cv(value):
                return value
            else:
                raise AVMItemNotInControlledVocabularyError(
                    "Item is not in the controlled vocabulary.")
        else:
            raise TypeError("{0:s} is not a string or unicode".format(self.tag))


class AVMStringCVCapitalize(AVMStringCV):

    def format_data(self, value):
        """
        Formats the data to be a capitalized string

        :return: String
        """
        return value.capitalize()


class AVMStringCVUpper(AVMStringCV):

    def format_data(self, value):
        """
        Formats the data to be an upper case string

        :return: String:
        """
        return value.upper()


class AVMLocalizedString(AVMString):

    def to_xml(self, parent, value):

        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))
        subelement = et.SubElement(element, "rdf:Alt")
        li = et.SubElement(subelement, "rdf:li")
        li.text = "%s" % value
        li.attrib['xml:lang'] = 'x-default'
        return element

# TODO: implement these
AVMDate = AVMString
AVMDateTime = AVMString


class AVMFloat(AVMData):
    """
    Data type for float fields
    """
    def check_data(self, value):
        """
        Checks that data can be represented as a number.

        :return: String (UTF-8)
        """
        if not value:
            return None

        try:
            value = float(value)
        except:
            raise TypeError(
                "Enter a value that can be represented as a number.")

        return value

    def to_xml(self, parent, value):
        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))
        element.text = "%.16f" % value
        return element


class AVMUnorderedList(AVMData):
    """
    Generic data type for lists (i.e xmp bag arrays)
    """
    def __init__(self, path, **kwargs):
        # Optional keyword arguments
        if 'length' in kwargs:
            self.length = kwargs['length']
        else:
            self.length = False

        if 'strict_length' in kwargs:
            self.strict_length = kwargs['strict_length']
        else:
            self.strict_length = False

        super(AVMUnorderedList, self).__init__(path, **kwargs)

    def check_length(self, values):
        """
        Checks the length of the Python List.

        :return: Boolean
        """
        if self.strict_length:
            if len(values) is self.length:
                return True
            else:
                return False
        elif self.length:
            if len(values) <= self.length:
                return True
            else:
                return False
        else:
            return True

    def check_data(self, values):
        """
        Checks that the data type is a Python List.  Calls check_length() first.

        .. todo :: Redo this function.  Implement the dash functionality only for ordered lists.

        :return: List (UTF-8 elements)
        """
        if not values:
            return None
        # Check data type
        if not isinstance(values, list):
            raise TypeError("Data needs to be a Python List.")

        # Check if all are None
        if all([v is None for v in values]):
            return None

        # Check length
        if not self.check_length(values):
            raise AVMListLengthError("Data is not the correct length.")

        # Convert to UTF-8
        checked_data = []
        length = 0

        for value in values:
            value = value
            length += len(value)
            if value is "":
                value = "-"
            checked_data.append(value)

        if len(set(checked_data)) == 1 and checked_data[0] == "-":
            checked_data = []

        return checked_data

    def to_xml(self, parent, values):

        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))

        subelement = et.SubElement(element, "rdf:Bag")

        for item in values:
            li = et.SubElement(subelement, "rdf:li")
            if type(item) is float:
                li.text = "%.16f" % item
            else:
                li.text = "%s" % item

        return element


class AVMUnorderedStringList(AVMUnorderedList):
    """
    Data type for an unordered list of strings
    """

    def check_data(self, values):
        """
        Check that the passed data is a Python List, and checks that the elements
        are strings or unicode.

        :return: List of strings (UTF-8)
        """
        # Check that data is a list
        if not isinstance(values, list):
            raise TypeError("Data needs to be a Python List.")

        # Check length
        if not self.check_length(values):
            raise AVMListLengthError("Data is not the correct length.")

        checked_data = []
        # Check data type in list
        for value in values:
            if (isinstance(value, basestring)):
                value = value
                checked_data.append(value)
            else:
                raise TypeError(
                    "Elements of list need to be string or unicode.")

        return checked_data

    def to_xml(self, parent, values):

        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))

        subelement = et.SubElement(element, "rdf:Bag")

        for item in values:
            li = et.SubElement(subelement, "rdf:li")
            li.text = "%s" % item

        return element


class AVMOrderedList(AVMUnorderedList):
    """
    Data type for ordered lists (i.e. seq arrays)
    """
    def to_xml(self, parent, values):

        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))

        subelement = et.SubElement(element, "rdf:Seq")

        for item in values:
            li = et.SubElement(subelement, "rdf:li")
            if type(item) is float:
                li.text = "%.16f" % item
            else:
                li.text = "%s" % item

        return element


class AVMOrderedListCV(AVMOrderedList, AVMStringCVCapitalize):
    """
    Data type for an ordered list constrained to a controlled vocabulary.
    """
    def __init__(self, path, cv, deprecated=False, **kwargs):

        self.namespace, self.tag = path.split(':')
        self.deprecated = deprecated
        self.controlled_vocabulary = cv

        # Optional keyword arguments
        if 'length' in kwargs:
            self.length = kwargs['length']
        else:
            self.length = False

        if 'strict_length' in kwargs:
            self.strict_length = kwargs['strict_length']
        else:
            self.strict_length = False

    def check_data(self, values):
        """
        Checks that the data is a list, elements are strings, and strings are in the specified controlled vocabulary.

        :return: List of CV-Strings (UTF-8)
        """
        # Check that data is a list
        if not isinstance(values, list):
            raise TypeError("Data needs to be a Python List.")

        # Check length
        if not self.check_length(values):
            raise AVMListLengthError("List is not the correct length.")

        checked_data = []
        # Check data type in list
        for value in values:
            if (isinstance(value, basestring)):
                value = value
                value = self.format_data(value)

                if self.check_cv(value):
                    checked_data.append(value)
                else:
                    raise AVMItemNotInControlledVocabularyError(
                        "Item is not in the controlled vocabulary.")
            else:
                if value is None:
                    checked_data.append("-")
                else:
                    raise TypeError(
                        "Elements of list need to be string or unicode.")

        if len(set(checked_data)) == 1 and checked_data[0] == "-":
            checked_data = []

        return checked_data


class AVMOrderedFloatList(AVMOrderedList):
    """
    Data type for ordered lists of floats.
    """
    def check_data(self, values):
        """
        Checks that the data is of the correct type, length and elements
        are strings able to be represented as floats.

        :return: List of strings (UTF-8)
        """
        checked_data = []

        if values:
            # Check type for list
            if not isinstance(values, list):
                raise TypeError("Data needs to be a list.")

            # Check length
            if not self.check_length(values):
                raise AVMListLengthError("Data is not the correct length.")

            # Check data type in list
            for value in values:
                if value.strip() == '-':
                    checked_data.append(None)
                else:
                    value = value
                    try:
                        checked_data.append(float(value))
                    except Exception:
                        raise TypeError("Enter a string that can be represented as a number.")

            if len(set(checked_data)) == 1 and checked_data[0] == "-":
                checked_data = []

        return checked_data

    def to_xml(self, parent, values):

        uri = reverse_namespaces[self.namespace]
        element = et.SubElement(parent, "{%s}%s" % (uri, self.tag))

        subelement = et.SubElement(element, "rdf:Bag")

        for item in values:
            li = et.SubElement(subelement, "rdf:li")
            if item is None:
                li.text = '-'
            else:
                li.text = "%.16f" % item

        return element


class AVMDateTimeList(AVMOrderedList):
    """
    Data type for lists composed of DateTime objects
    """
    def check_data(self, values):
        """
        Checks that the data passed is a Python List,
        and that the elements are Date or Datetime objects.

        :return: List of Datetime objects in ISO format (i.e. Strings encoded as UTF-8)
        """
        if not isinstance(values, list):
            raise TypeError("Data needs to be a list.")

        if not self.check_length(values):
            raise AVMListLengthError("Data is not the correct length.")

        checked_data = []
        # Check data type in list
        for value in values:
            if value:
                if (isinstance(value, datetime.date) or isinstance(value, datetime.datetime)):
                    value = value.isoformat()
                    checked_data.append(value)
                elif isinstance(value, basestring):
                    value = value
                    checked_data.append(value)
                else:
                    raise TypeError("Elements of the list need to be a Python Date or Datetime object.")
            else:
                checked_data.append("-")

        if len(set(checked_data)) == 1 and checked_data[0] == "-":
            checked_data = []

        return checked_data



from __future__ import print_function, division

import struct
import warnings

from .jpeg import is_jpeg, JPEGFile, JPEGSegment
from .png import is_png, PNGFile, PNGChunk


def embed_xmp(image_in, image_out, xmp_packet):

    if is_jpeg(image_in):

        # Check length
        if len(xmp_packet) >= 65503:
            raise Exception("XMP packet is too long to embed in JPG file")

        # XMP segment
        xmp_segment = JPEGSegment()

        # APP1 marker
        xmp_segment.bytes = b"\xff\xe1"

        # Length of XMP packet + 2 + 29
        xmp_segment.bytes += struct.pack('>H', len(xmp_packet) + 29 + 2)

        # XMP Namespace URI (NULL-terminated)
        xmp_segment.bytes += b"http://ns.adobe.com/xap/1.0/\x00"

        # XMP packet
        xmp_segment.bytes += xmp_packet

        # Read in input file
        jpeg_file = JPEGFile.read(image_in)

        # Check if there is already XMP data in the file
        existing = []
        for segment in jpeg_file.segments:
            if segment.type == 'APP1':
                if segment.bytes[4:32] == b'http://ns.adobe.com/xap/1.0/':
                    existing.append(segment)
        if existing:
            warnings.warn("Discarding existing XMP packet from JPEG file")
            for e in existing:
                jpeg_file.segments.remove(e)

        # Position at which to insert the packet
        markers = [x.type for x in jpeg_file.segments]

        if 'APP1' in markers:  # Put it after existing APP1
            index = markers.index('APP1') + 1
        elif 'APP0' in markers:  # Put it after existing APP0
            index = markers.index('APP0') + 1
        elif 'SOF' in markers:
            index = markers.index('SOF')
        else:
            raise ValueError("Could not find SOF marker")

        # Insert segment into JPEG file
        jpeg_file.segments.insert(index, xmp_segment)
        jpeg_file.write(image_out)

    elif is_png(image_in):

        xmp_chunk = PNGChunk()

        # Keyword
        xmp_chunk.data = b'XML:com.adobe.xmp'

        # Null separator
        xmp_chunk.data += b'\x00'

        # Compression flag
        xmp_chunk.data += b'\x00'

        # Compression method
        xmp_chunk.data += b'\x00'

        # Null separator
        xmp_chunk.data += b'\x00'

        # Null separator
        xmp_chunk.data += b'\x00'

        # Text
        xmp_chunk.data += xmp_packet

        # Set type
        xmp_chunk.type = b'iTXt'

        # Read in input file
        png_file = PNGFile.read(image_in)

        # Check if there is already XMP data in the file
        existing = []
        for chunk in png_file.chunks:
            if chunk.type == b'iTXt':
                if chunk.data.startswith(b'XML:com.adobe.xmp'):
                    existing.append(chunk)
        if existing:
            warnings.warn("Discarding existing XMP packet from PNG file")
            for e in existing:
                png_file.chunks.remove(e)

        # Insert chunk into PNG file
        png_file.chunks.insert(1, xmp_chunk)
        png_file.write(image_out)

    else:

        raise Exception("Only JPG and PNG files are supported at this time")



# -*- coding: utf-8 -*-

from __future__ import print_function, division


class AVMListLengthError(Exception):
    """
    Raised when a list is not the correct length
    """
    pass


class AVMItemNotInControlledVocabularyError(Exception):
    """
    Raised when an string is not in a required controlled vocabulary
    """
    pass


class AVMEmptyValueError(Exception):
    """
    Raised when a list is given with no relevant data
    """
    pass


class NoXMPPacketFound(Exception):
    """
    Raised when no XMP packet is found in a file
    """
    pass



import pytest
import warnings
warnings.filterwarnings('always')

from ..avm import AVM, AVMContainer


@pytest.mark.parametrize('version', [1.1, 1.2])
def test_specs(version):

    a = AVM(version=version)

    # Creator Metadata
    a.Creator = "PyAVM"
    a.CreatorURL = "http://www.github.com"
    a.Contact.Name = ["Thomas Robitaille"]
    a.Contact.Email = "thomas.robitaille@gmail.com"
    a.Contact.Address = "None of your business"
    a.Contact.Telephone = "I think we're getting a little too personal"
    a.Contact.City = "Heidelberg"
    a.Contact.StateProvince = "Baden-Wuttemburg"
    a.Contact.PostalCode = "What could you possibly need this for?"
    a.Contact.Country = "Germany"
    a.Rights = "Wrongs"

    # Content Metadata
    a.Title = "A very thorough test"
    a.Headline = "What I said above"
    a.Description = "Um, I guess there's not much more to say about this!"
    a.Subject.Category = ["Tests"]
    a.Subject.Name = ["PyAVM"]
    a.Distance = ["3"]
    a.Distance.Notes = "Not much to say, really"
    a.ReferenceURL = "http://www.github.com"
    a.Credit = "Me"
    a.Date = "10 April 2013"
    a.ID = "123123123"
    a.Type = "Simulation"
    a.Image.ProductQuality = "Moderate"

    # Observation Metadata
    a.Facility = ["Python"]
    a.Instrument = ["CPython"]
    a.Spectral.ColorAssignment = ["Purple"]
    a.Spectral.Band = ["Optical"]
    a.Spectral.Bandpass = ["Arbitrary"]
    a.Spectral.CentralWavelength = [5.]
    a.Spectral.Notes = "Still testing"
    a.Temporal.StartTime = ["5 Feb 2011"]
    a.Temporal.IntegrationTime = [4.4]
    a.DatasetID = ["12421412"]

    # Coordinate Metadata
    a.Spatial.CoordinateFrame = "GAL"
    a.Spatial.Equinox = '2000'
    a.Spatial.ReferenceValue = [33.3, 44.4]
    a.Spatial.ReferenceDimension = [300, 400]
    a.Spatial.ReferencePixel = [2., 3.]
    a.Spatial.Scale = [0.2, 0.3]
    a.Spatial.Rotation = 122.
    a.Spatial.CoordsystemProjection = "CAR"
    a.Spatial.Quality = "Full"
    a.Spatial.Notes = "Not much to say"
    a.Spatial.FITSheader = "SIMPLE = T"
    a.Spatial.CDMatrix = [3.4, 3.3, 5.5, 2.1]

    # Publisher Metadata
    a.Publisher = "Tom"
    a.PublisherID = "125521"
    a.ResourceID = "3995611"
    a.ResourceURL = "http://www.github.com"
    a.RelatedResources = ["Testing", "Python", "PyAVM"]
    a.MetadataDate = "20 April 2013"

    # FITS Liberator Metadata

    a.FL.BackgroundLevel = [3.4]
    a.FL.BlackLevel = [4.4]
    a.FL.ScaledPeakLevel = [5.5]
    a.FL.PeakLevel = [10.2]
    a.FL.WhiteLevel = [11.3]
    a.FL.ScaledBackgroundLevel = [4.5]
    a.FL.StretchFunction = ['Log']

    # Spec-dependent keywords
    if version == 1.1:
        with pytest.raises(AttributeError) as exc:
            a.ProposalID = ["12421412"]
        assert exc.value.args[0] == "ProposalID is not a valid AVM group or tag in the 1.1 standard"
        with pytest.raises(AttributeError) as exc:
            a.PublicationID = ['799292']
        assert exc.value.args[0] == "PublicationID is not a valid AVM group or tag in the 1.1 standard"
    else:
        a.ProposalID = ["12421412"]
        a.PublicationID = ['799292']

    x = a.to_xml()

    b = AVM.from_xml(x)

    for key in a._items:
        if isinstance(a._items[key], AVMContainer):
            for subkey in a._items[key]._items:
                assert a._items[key]._items[subkey] == b._items[key]._items[subkey]
        else:
            assert a._items[key] == b._items[key]


def test_warning():

    # Start of with a version=1.2 AVM object
    a = AVM(version=1.2)
    a.ProposalID = ["25661"]

    # Then change to version=1.1, which doesn't contain ProposalID
    with warnings.catch_warnings(record=True) as w:
        a.MetadataVersion = 1.1
        assert len(w) == 1
        assert str(w[0].message) == "ProposalID is not defined in format specification 1.1 and will be deleted"

    try:
        a.ProposalID = ["44663"]
    except AttributeError as exc:
        assert exc.args[0] == "ProposalID is not a valid AVM group or tag in the 1.1 standard"



# Pure-python JPEG parser
# Copyright (c) 2013 Thomas P. Robitaille

from __future__ import print_function, division

import struct

# Define common markers

MARKERS = {}
MARKERS[b'\xd8'] = 'SOI'
MARKERS[b'\xc0'] = 'SOF0'
MARKERS[b'\xc2'] = 'SOF2'
MARKERS[b'\xc4'] = 'DHT'
MARKERS[b'\xdb'] = 'DQT'
MARKERS[b'\xdd'] = 'DRI'
MARKERS[b'\xda'] = 'SOS'
MARKERS[b'\xd0'] = 'RST0'
MARKERS[b'\xd1'] = 'RST1'
MARKERS[b'\xd2'] = 'RST2'
MARKERS[b'\xd3'] = 'RST3'
MARKERS[b'\xd4'] = 'RST4'
MARKERS[b'\xd5'] = 'RST5'
MARKERS[b'\xd6'] = 'RST6'
MARKERS[b'\xd7'] = 'RST7'
MARKERS[b'\xe0'] = 'APP0'
MARKERS[b'\xe1'] = 'APP1'
MARKERS[b'\xe2'] = 'APP2'
MARKERS[b'\xe3'] = 'APP3'
MARKERS[b'\xe4'] = 'APP4'
MARKERS[b'\xe5'] = 'APP5'
MARKERS[b'\xe6'] = 'APP6'
MARKERS[b'\xe7'] = 'APP7'
MARKERS[b'\xe8'] = 'APP8'
MARKERS[b'\xe9'] = 'APP9'
MARKERS[b'\xfe'] = 'COM'
MARKERS[b'\xd9'] = 'EOI'

# Define some markers which are always followed by variable-length data

VARIABLE = ['APP0', 'APP1', 'APP2', 'APP3', 'APP4',
            'APP5', 'APP6', 'APP7', 'APP8', 'APP9']


def is_jpeg(filename):
    with open(filename, 'rb') as f:
        return f.read(4) == b'\xff\xd8\xff\xe0'


class JPEGSegment(object):

    @classmethod
    def from_bytes(cls, bytes):
        self = cls()
        if bytes[1:2] in MARKERS:
            self.type = MARKERS[bytes[1:2]]
        else:
            self.type = "UNKNOWN"
        self.bytes = bytes
        return self

    def write(self, fileobj):
        fileobj.write(self.bytes)


class JPEGFile(object):

    @classmethod
    def read(cls, filename):

        fileobj = open(filename, 'rb')
        contents = fileobj.read()
        fileobj.close()

        self = cls()

        # Search for all starts of segments. Segments all start with 0xff, but
        # we should ignore 0xff instances that are followed by 0x00.

        self.segments = []

        start = end = 0
        while True:
            if contents[start + 1:start + 2] in MARKERS and MARKERS[contents[start + 1:start + 2]] in VARIABLE:
                end = contents.find(b'\xff', end + 4)
            else:
                end = contents.find(b'\xff', end + 1)
            if end == -1 or contents[end + 1:end + 2] not in [b'\xff', b'\x00']:
                if end == -1:
                    end = len(contents) + 1
                self.segments.append(JPEGSegment.from_bytes(contents[start:end]))
                if end == len(contents) + 1:
                    break
                else:
                    start = end

        if self.segments[0].type != 'SOI':
            raise ValueError("Image did not start with SOI but with {0}".format(self.segments[0].type))

        if self.segments[-1].type != 'EOI':
            raise ValueError("Image did not end with EOI but with {0}".format(self.segments[-1].type))

        return self

    def write(self, filename):

        fileobj = open(filename, 'wb')

        for segment in self.segments:
            segment.write(fileobj)

        fileobj.close()



from __future__ import print_function, division

import os
import glob
import warnings

import pytest

try:
    from PIL import Image
except ImportError:
    try:
        import Image
    except ImportError:
        pytest.skip()

try:
    import numpy as np
except ImportError:
    pytest.skip()

from .. import AVM

ROOT = os.path.dirname(os.path.abspath(__file__))

XML_FILES = glob.glob(os.path.join(ROOT, 'data', '*.xml'))


@pytest.mark.parametrize('xml_file', XML_FILES)
def test_io_png(tmpdir, xml_file):
    avm = AVM.from_xml_file(xml_file)
    filename_in = tmpdir.join('test_in.png').strpath
    filename_out = tmpdir.join('test_out.png').strpath
    i = Image.fromarray(np.ones((16, 16), dtype=np.uint8))
    i.save(filename_in)
    avm.embed(filename_in, filename_out, verify=True)


@pytest.mark.parametrize('xml_file', XML_FILES)
def test_io_jpeg(tmpdir, xml_file):
    avm = AVM.from_xml_file(xml_file)
    filename_in = tmpdir.join('test_in.jpg').strpath
    filename_out = tmpdir.join('test_out.jpg').strpath
    i = Image.fromarray(np.ones((16, 16), dtype=np.uint8))
    i.save(filename_in)
    avm.embed(filename_in, filename_out, verify=True)


@pytest.mark.parametrize('xml_file', XML_FILES)
def test_io_png_repeat(tmpdir, xml_file):
    warnings.simplefilter('always')
    avm = AVM.from_xml_file(xml_file)
    filename_in = tmpdir.join('test_in.png').strpath
    filename_out_1 = tmpdir.join('test_out_1.png').strpath
    filename_out_2 = tmpdir.join('test_out_2.png').strpath
    i = Image.fromarray(np.ones((16, 16), dtype=np.uint8))
    i.save(filename_in)
    with warnings.catch_warnings(record=True) as w:
        avm.embed(filename_in, filename_out_1, verify=True)
        messages = [str(x.message) for x in w]
        assert 'Discarding existing XMP packet from PNG file' not in messages
    with warnings.catch_warnings(record=True) as w:
        avm.embed(filename_out_1, filename_out_2, verify=True)
        messages = [str(x.message) for x in w]
        assert 'Discarding existing XMP packet from PNG file' in messages


@pytest.mark.parametrize('xml_file', XML_FILES)
def test_io_jpeg_repeat(tmpdir, xml_file):
    warnings.simplefilter('always')
    avm = AVM.from_xml_file(xml_file)
    filename_in = tmpdir.join('test_in.jpg').strpath
    filename_out_1 = tmpdir.join('test_out_1.jpg').strpath
    filename_out_2 = tmpdir.join('test_out_2.jpg').strpath
    i = Image.fromarray(np.ones((16, 16), dtype=np.uint8))
    i.save(filename_in)
    with warnings.catch_warnings(record=True) as w:
        avm.embed(filename_in, filename_out_1, verify=True)
        messages = [str(x.message) for x in w]
        assert 'Discarding existing XMP packet from JPEG file' not in messages
    with warnings.catch_warnings(record=True) as w:
        avm.embed(filename_out_1, filename_out_2, verify=True)
        messages = [str(x.message) for x in w]
        assert 'Discarding existing XMP packet from JPEG file' in messages



from __future__ import print_function, division

import os
import glob

import pytest

from ..avm import AVM, NoSpatialInformation

ROOT = os.path.dirname(os.path.abspath(__file__))

XML_FILES = glob.glob(os.path.join(ROOT, 'data', '*.xml'))


@pytest.mark.parametrize('filename', XML_FILES)
def test_parse(filename):
    AVM.from_xml_file(filename)


@pytest.mark.parametrize('filename', XML_FILES)
def test_to_xml(filename):
    a = AVM.from_xml_file(filename)
    a.to_xml()


@pytest.mark.parametrize('filename', XML_FILES)
def test_to_xmp(filename):
    a = AVM.from_xml_file(filename)
    a.to_xmp()


NO_WCS = [os.path.join(ROOT, 'data', 'heic0409a.xml'),
          os.path.join(ROOT, 'data', 'sig05-021-alpha.xml'),
          os.path.join(ROOT, 'data', 'ssc2004-06a1-alpha.xml'),
          os.path.join(ROOT, 'data', 'ssc2004-06b1-alpha.xml')]

XML_FILES_WCS = [x for x in XML_FILES if x not in NO_WCS]


@pytest.mark.parametrize('filename', XML_FILES_WCS)
def test_to_wcs(filename):
    a = AVM.from_xml_file(filename)
    a.to_wcs()


@pytest.mark.parametrize('filename', XML_FILES_WCS)
def test_to_wcs_target_image(filename, tmpdir):
    from PIL import Image
    image = Image.fromstring(data=b"1111", size=(2,2), mode="L")
    image_file = tmpdir.join('test.png').strpath
    image.save(image_file)
    image.close()
    a = AVM.from_xml_file(filename)
    a.Spatial.ReferenceDimension = (30, 30)
    a.to_wcs(target_image=image_file)



@pytest.mark.parametrize('filename', NO_WCS)
def test_to_wcs_nowcs(filename):
    a = AVM.from_xml_file(filename)
    with pytest.raises(NoSpatialInformation):
        a.to_wcs()



# -*- coding: utf-8 -*-

from __future__ import print_function, division

"""
Storage for controlled vocabulary fields
"""

TYPE_CHOICES = [
    'Observation',
    'Artwork',
    'Photographic',
    'Planetary',
    'Simulation',
    'Collage',
    'Chart'
]

IMAGE_PRODUCT_QUALITY_CHOICES = [
    'Good',
    'Moderate',
    'Poor'
]

SPECTRAL_COLOR_ASSIGNMENT_CHOICES = [
    'Purple',
    'Blue',
    'Cyan',
    'Green',
    'Yellow',
    'Orange',
    'Red',
    'Magenta',
    'Grayscale',
    'Pseudocolor',
    'Luminosity',
]

SPECTRAL_BAND_CHOICES = [
    'Radio',
    'Millimeter',
    'Infrared',
    'Optical',
    'Ultraviolet',
    'X-ray',
    'Gamma-ray'
]

SPATIAL_COORDINATE_FRAME_CHOICES = [
    'ICRS',
    'FK5',
    'FK4',
    'ECL',
    'GAL',
    'SGAL',
]

SPATIAL_EQUINOX_CHOICES = [
    'J2000',
    'B1950',
]

SPATIAL_COORDSYSTEM_PROJECTION_CHOICES = [
    'TAN',
    'SIN',
    'ARC',
    'AIT',
    'CAR',
    'CEA',
]

SPATIAL_QUALITY_CHOICES = [
    'Full',
    'Position',
]

METADATA_VERSION_CHOICES = [
    1.2,
    1.1,
    1.0,
]


from __future__ import print_function, division

try:
    unicode
except:
    basestring = unicode = str

import os
import pytest

from ..avm import AVM, NoSpatialInformation

ROOT = os.path.dirname(os.path.abspath(__file__))


def test_from_header():
    from astropy.io import fits
    header = fits.Header.fromtextfile(os.path.join(ROOT, 'data', 'example_header.hdr'))
    a = AVM.from_header(header)
    assert isinstance(a.Spatial.FITSheader, basestring)
    assert a.Spatial.FITSheader == header
    # assert a.Spatial.Equinox == 2000.  # returns NaN at the moment
    assert a.Spatial.CoordsystemProjection == 'CAR'
    assert a.Spatial.ReferenceDimension[0] == 599
    assert a.Spatial.ReferenceDimension[1] == 599
    assert a.Spatial.ReferenceValue[0] == 0.
    assert a.Spatial.ReferenceValue[1] == 0.
    assert a.Spatial.ReferencePixel[0] == 299.628
    assert a.Spatial.ReferencePixel[1] == 299.394
    assert a.Spatial.Scale[0] == -0.001666666707
    assert a.Spatial.Scale[1] == +0.001666666707
    assert a.Spatial.Quality == 'Full'


def test_from_header_cd():
    from astropy.io import fits
    header = fits.Header.fromtextfile(os.path.join(ROOT, 'data', 'example_header.hdr'))
    header['CD1_1'] = header.pop('CDELT1')
    header['CD2_2'] = header.pop('CDELT2')
    a = AVM.from_header(header)
    assert isinstance(a.Spatial.FITSheader, basestring)
    assert a.Spatial.FITSheader == header
    # assert a.Spatial.Equinox == 2000.  # returns NaN at the moment
    assert a.Spatial.CoordsystemProjection == 'CAR'
    assert a.Spatial.ReferenceDimension[0] == 599
    assert a.Spatial.ReferenceDimension[1] == 599
    assert a.Spatial.ReferenceValue[0] == 0.
    assert a.Spatial.ReferenceValue[1] == 0.
    assert a.Spatial.ReferencePixel[0] == 299.628
    assert a.Spatial.ReferencePixel[1] == 299.394
    assert a.Spatial.Scale[0] == -0.001666666707
    assert a.Spatial.Scale[1] == +0.001666666707
    assert a.Spatial.Quality == 'Full'


def test_wcs_1():
    from astropy.io import fits
    header = fits.Header.fromtextfile(os.path.join(ROOT, 'data', 'example_header.hdr'))
    a = AVM.from_header(header)
    b = AVM.from_wcs(a.to_wcs(), shape=(header['NAXIS2'], header['NAXIS1']))
    # assert a.Spatial.Equinox == b.Spatial.Equinox  # returns NaN at the moment
    assert a.Spatial.CoordsystemProjection == b.Spatial.CoordsystemProjection
    assert a.Spatial.ReferenceDimension[0] == b.Spatial.ReferenceDimension[0]
    assert a.Spatial.ReferenceDimension[1] == b.Spatial.ReferenceDimension[1]
    assert a.Spatial.ReferenceValue[0] == b.Spatial.ReferenceValue[0]
    assert a.Spatial.ReferenceValue[1] == b.Spatial.ReferenceValue[1]
    assert a.Spatial.ReferencePixel[0] == b.Spatial.ReferencePixel[0]
    assert a.Spatial.ReferencePixel[1] == b.Spatial.ReferencePixel[1]
    assert a.Spatial.Scale[0] == b.Spatial.Scale[0]
    assert a.Spatial.Scale[1] == b.Spatial.Scale[1]
    assert a.Spatial.Quality == b.Spatial.Quality


def test_wcs_2():
    from astropy.io import fits
    from astropy.wcs import WCS
    header = fits.Header.fromtextfile(os.path.join(ROOT, 'data', 'example_header.hdr'))
    a = WCS(header)
    b = AVM.from_wcs(a).to_wcs()
    # assert a.wcs.equinox == b.wcs.equinox
    assert a.wcs.ctype[0] == b.wcs.ctype[0]
    assert a.wcs.ctype[1] == b.wcs.ctype[1]
    assert a.wcs.crval[0] == b.wcs.crval[0]
    assert a.wcs.crval[1] == b.wcs.crval[1]
    assert a.wcs.crpix[0] == b.wcs.crpix[0]
    assert a.wcs.crpix[1] == b.wcs.crpix[1]
    assert a.wcs.cdelt[0] == b.wcs.cdelt[0]
    assert a.wcs.cdelt[1] == b.wcs.cdelt[1]
    assert a.wcs.crota[0] == b.wcs.crota[0]
    assert a.wcs.crota[1] == b.wcs.crota[1]
    assert a.wcs.radesys == b.wcs.radesys



#!/usr/bin/env python

from distutils.core import setup

try:  # Python 3.x
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:  # Python 2.x
    from distutils.command.build_py import build_py

version = '0.9.2'

setup(name='PyAVM',
      version=version,
      description='Simple pure-python AVM meta-data handling',
      author='Thomas Robitaille',
      author_email='thomas.robitaille@gmail.com',
      license='MIT',
      url='http://astrofrog.github.io/pyavm/',
      packages=['pyavm', 'pyavm.tests'],
      package_data={'pyavm.tests':['data/*.xml', 'data/*.hdr']},
      provides=['pyavm'],
      cmdclass={'build_py': build_py},
      keywords=['Scientific/Engineering'],
      long_description="PyAVM is a module to represent, read, and write metadata following the `Astronomy Visualization Metadata <http://www.virtualastronomy.org/avm_metadata.php>`_ (AVM) standard.",
      classifiers=[
      "Development Status :: 4 - Beta",
      "Programming Language :: Python",
      "License :: OSI Approved :: MIT License",
      ],
      )



from __future__ import print_function, division

import struct
import warnings

from .jpeg import is_jpeg, JPEGFile, JPEGSegment
from .png import is_png, PNGFile, PNGChunk
from .exceptions import NoXMPPacketFound

def extract_xmp(image):

    if is_jpeg(image):

        # Read in input file
        jpeg_file = JPEGFile.read(image)

        # Loop through segments and search for XMP packet
        for segment in jpeg_file.segments:
            if segment.type == 'APP1':
                if segment.bytes[4:32] == b'http://ns.adobe.com/xap/1.0/':
                    return segment.bytes[32:]

        # No XMP data was found
        raise NoXMPPacketFound("No XMP packet present in file")

    elif is_png(image):

        # Read in input file
        png_file = PNGFile.read(image)

        # Loop through chunks and search for XMP packet
        for chunk in png_file.chunks:
            if chunk.type == 'iTXt':
                if chunk.data.startswith(b'XML:com.adobe.xmp'):
                    return chunk.data[22:]

        # No XMP data was found
        raise NoXMPPacketFound("No XMP packet present in file")

    else:

        warnings.warn("Only PNG and JPEG files can be properly parsed - scanning file contents for XMP packet")

        with open(image, 'rb') as fileobj:
            contents = fileobj.read()

        try:
            start = contents.index(b"<?xpacket begin=")
            end = contents.index(b"</x:xmpmeta>") + 12
        except ValueError:
            raise NoXMPPacketFound("No XMP packet present in file")

        return contents[start:end]



# Pure-python PNG parser
# Copyright (c) 2013 Thomas P. Robitaille

from __future__ import print_function, division

import struct

PNG_SIGNATURE = b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'


def is_png(filename):
    with open(filename, 'rb') as f:
        return f.read(8) == b'\x89\x50\x4e\x47\x0d\x0a\x1a\x0a'


class PNGChunk(object):

    @classmethod
    def read(cls, fileobj):

        self = cls()

        # Read in chunk length
        length = struct.unpack('>I', fileobj.read(4))[0]

        # Read in chunk type
        self.type = fileobj.read(4)

        # Read in data
        self.data = fileobj.read(length)

        # Read in CRC
        crc = struct.unpack('>I', fileobj.read(4))[0]

        # Check that the CRC matches the actual one
        if crc != self.crc:
            raise ValueError("CRC ({0}) does not match advertised ({1})".format(self.crc, crc))

        if length != self.length:
            raise ValueError("Dynamic length ({0}) does not match original length ({1})".format(self.length, length))

        return self

    def write(self, fileobj):

        # Write length
        fileobj.write(struct.pack('>I', self.length))

        # Write type
        fileobj.write(self.type)

        # Write data
        fileobj.write(self.data)

        # Write CRC
        fileobj.write(struct.pack('>I', self.crc))

    @property
    def crc(self):
        # Note from Python docs: "To generate the same numeric value across all
        # Python versions and platforms use crc32(data) & 0xffffffff. If you
        # are only using the checksum in packed binary format this is not
        # necessary as the return value is the correct 32bit binary
        # representation regardless of sign."
        # This is indeed true, I see different values in Python 2 and 3.
        from zlib import crc32
        return crc32(self.type + self.data) & 0xffffffff

    @property
    def length(self):
        return len(self.data)


class PNGFile(object):

    @classmethod
    def read(cls, filename):

        fileobj = open(filename, 'rb')

        self = cls()

        # Read signature
        sig = fileobj.read(8)

        if sig != PNG_SIGNATURE:
            raise ValueError("Signature ({0}) does match expected ({1})".format(sig, PNG_SIGNATURE))

        self.chunks = []

        while True:
            chunk = PNGChunk.read(fileobj)
            self.chunks.append(chunk)
            if chunk.type == b'IEND':
                break

        fileobj.close()

        return self

    def write(self, filename):

        fileobj = open(filename, 'wb')

        fileobj.write(PNG_SIGNATURE)

        for chunk in self.chunks:
            chunk.write(fileobj)

        fileobj.close()
